# pylint: disable=missing-docstring
from collections import OrderedDict

from guardian.shortcuts import assign_perm
from rest_framework.test import APIRequestFactory

from resolwe.flow.models import Data, DescriptorSchema, Process
from resolwe.flow.serializers import DataSerializer
from resolwe.test import TestCase


class ResolweDictRelatedFieldTest(TestCase):
    def setUp(self):
        super().setUp()

        self.process = Process.objects.create(
            slug="test-process",
            contributor=self.contributor,
        )
        assign_perm("view_process", self.user, self.process)

        self.descriptor_schema1 = DescriptorSchema.objects.create(
            slug="test-schema",
            contributor=self.contributor,
            version="1.0.0",
        )
        assign_perm("view_descriptorschema", self.user, self.descriptor_schema1)

        self.descriptor_schema2 = DescriptorSchema.objects.create(
            slug="test-schema",
            contributor=self.contributor,
            version="2.0.0",
        )
        assign_perm("view_descriptorschema", self.user, self.descriptor_schema2)

        self.descriptor_schema3 = DescriptorSchema.objects.create(
            slug="test-schema",
            contributor=self.contributor,
            version="3.0.0",
        )

        self.factory = APIRequestFactory()

    def test_to_internal_value(self):
        request = self.factory.get("/")
        request.user = self.user
        request.query_params = {}
        data = {
            "contributor": self.user.pk,
            "process": {"slug": "test-process"},
            "descriptor_schema": {"slug": "test-schema"},
        }

        serializer = DataSerializer(data=data, context={"request": request})
        # is_valid() needs to be called before accessing ``validated_data``
        serializer.is_valid()
        # Check that descriptor schmena with highest version & view permission is used:
        self.assertEqual(
            serializer.validated_data["descriptor_schema"], self.descriptor_schema2
        )

    def test_to_representation(self):
        request = self.factory.get("/")
        request.user = self.user
        request.query_params = {}

        data = Data.objects.create(
            contributor=self.user,
            process=self.process,
            descriptor_schema=self.descriptor_schema1,
        )

        serializer = DataSerializer(data, context={"request": request})
        self.assertEqual(serializer.data["process"]["id"], self.process.pk)

        # Check that descriptor_schema is properly hydrated (but remove
        # values that are not deterministic from the checking procedure)
        descriptor_schema_hydrated = serializer.data["descriptor_schema"]
        for key in ["created", "modified", "id"]:
            self.assertTrue(key in descriptor_schema_hydrated)
            descriptor_schema_hydrated.pop(key)
        descriptor_schema_hydrated.get("contributor", {}).pop("id")
        self.assertDictEqual(
            descriptor_schema_hydrated,
            {
                "slug": "test-schema",
                "version": "1.0.0",
                "name": "",
                "description": "",
                "schema": [],
                "contributor": OrderedDict(
                    [
                        ("first_name", "Joe"),
                        ("last_name", "Miller"),
                        ("username", "contributor"),
                    ]
                ),
            },
        )



"""Resolwe entity serializer."""
from resolwe.flow.models import Collection, Entity

from .collection import BaseCollectionSerializer
from .fields import DictRelatedField


class EntitySerializer(BaseCollectionSerializer):
    """Serializer for Entity."""

    collection = DictRelatedField(
        queryset=Collection.objects.all(),
        serializer=BaseCollectionSerializer,
        allow_null=True,
        required=False,
        write_permission="edit",
    )

    class Meta(BaseCollectionSerializer.Meta):
        """EntitySerializer Meta options."""

        model = Entity
        fields = BaseCollectionSerializer.Meta.fields + (
            "collection",
            "duplicated",
            "type",
        )

    def update(self, instance, validated_data):
        """Update."""
        source_collection = instance.collection
        instance = super().update(instance, validated_data)
        instance.move_to_collection(source_collection, instance.collection)

        return instance



"""Resolwe collection model."""
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.db import models, transaction

from .base import BaseModel, BaseQuerySet
from .utils import DirtyError, bulk_duplicate, validate_schema


class BaseCollection(BaseModel):
    """Template for Postgres model for storing a collection."""

    class Meta(BaseModel.Meta):
        """BaseCollection Meta options."""

        abstract = True

    #: detailed description
    description = models.TextField(blank=True)

    settings = models.JSONField(default=dict)

    #: collection descriptor schema
    descriptor_schema = models.ForeignKey(
        "flow.DescriptorSchema", blank=True, null=True, on_delete=models.PROTECT
    )

    #: collection descriptor
    descriptor = models.JSONField(default=dict)

    #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
    descriptor_dirty = models.BooleanField(default=False)

    #: tags for categorizing objects
    tags = ArrayField(models.CharField(max_length=255), default=list)

    #: field used for full-text search
    search = SearchVectorField(null=True)

    def save(self, *args, **kwargs):
        """Perform descriptor validation and save object."""
        if self.descriptor_schema:
            try:
                validate_schema(self.descriptor, self.descriptor_schema.schema)
                self.descriptor_dirty = False
            except DirtyError:
                self.descriptor_dirty = True
        elif self.descriptor and self.descriptor != {}:
            raise ValueError(
                "`descriptor_schema` must be defined if `descriptor` is given"
            )

        super().save()


class CollectionQuerySet(BaseQuerySet):
    """Query set for ``Collection`` objects."""

    @transaction.atomic
    def duplicate(self, contributor):
        """Duplicate (make a copy) ``Collection`` objects."""
        return bulk_duplicate(collections=self, contributor=contributor)


class Collection(BaseCollection):
    """Postgres model for storing a collection."""

    class Meta(BaseCollection.Meta):
        """Collection Meta options."""

        permissions = (
            ("view_collection", "Can view collection"),
            ("edit_collection", "Can edit collection"),
            ("share_collection", "Can share collection"),
            ("owner_collection", "Is owner of the collection"),
        )

        indexes = [
            models.Index(name="idx_collection_name", fields=["name"]),
            GinIndex(
                name="idx_collection_name_trgm",
                fields=["name"],
                opclasses=["gin_trgm_ops"],
            ),
            models.Index(name="idx_collection_slug", fields=["slug"]),
            GinIndex(name="idx_collection_tags", fields=["tags"]),
            GinIndex(name="idx_collection_search", fields=["search"]),
        ]

    #: manager
    objects = CollectionQuerySet.as_manager()

    #: duplication date and time
    duplicated = models.DateTimeField(blank=True, null=True)

    def is_duplicate(self):
        """Return True if collection is a duplicate."""
        return bool(self.duplicated)

    def duplicate(self, contributor):
        """Duplicate (make a copy)."""
        return bulk_duplicate(
            collections=self._meta.model.objects.filter(pk=self.pk),
            contributor=contributor,
        )[0]



# Generated by Django 2.2.10 on 2020-05-12 08:44
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

from django.db import migrations


from resolwe.storage.connectors import connectors


def process_storage_location(file_storage, best_storage_location):
    connector = connectors[best_storage_location.connector_name].duplicate()
    base_url = Path(best_storage_location.url)
    for referenced_path in file_storage.files.exclude(path__endswith="/"):
        url = base_url / referenced_path.path
        hashes = connector.get_hashes(url, ["md5", "crc32c", "awss3etag"])
        referenced_path.md5 = hashes["md5"]
        referenced_path.crc32c = hashes["crc32c"]
        referenced_path.awss3etag = hashes["awss3etag"]
        referenced_path.save()


def calculate_hashes(apps, schema_editor):
    """Calculate hashes for existing ReferencedPaths."""
    FileStorage = apps.get_model("storage", "FileStorage")

    with ThreadPoolExecutor(max_workers=50) as executor:
        for file_storage in FileStorage.objects.all():
            storage_locations = file_storage.storage_locations.filter(status="OK")
            # Do not calculate hash when no location with status OK exists.
            if storage_locations.count() == 0:
                continue
            best_storage_location = storage_locations.first()
            best_priority = connectors[best_storage_location.connector_name].priority
            for storage_location in storage_locations:
                priority = connectors[storage_location.connector_name].priority
                if priority < best_priority:
                    best_storage_location = storage_location
                    best_priority = priority
            storage_location = best_storage_location
            executor.submit(
                process_storage_location, file_storage, best_storage_location
            )


class Migration(migrations.Migration):
    dependencies = [
        ("storage", "0003_add_hash_fields_to_referenced_path"),
    ]

    operations = [
        migrations.RunPython(calculate_hashes),
    ]



""".. Ignore pydocstyle D400.

==========
Base Class
==========

.. autoclass:: resolwe.flow.executors.run.BaseFlowExecutor
    :members:

"""
# pylint: disable=logging-format-interpolation
import logging
import sys
from pathlib import Path
from typing import Dict, Tuple

from .global_settings import LOCATION_SUBPATH, PROCESS, SETTINGS
from .zeromq_utils import ZMQCommunicator

# TODO: update requirements!!!!
# NOTE: If the imports here are changed, the executors' requirements.txt
# file must also be updated accordingly.
logger = logging.getLogger(__name__)


class BaseFlowExecutor:
    """Represents a workflow executor."""

    # Name has to be overriden in inherited classes.
    name = "base"

    def __init__(
        self,
        data_id: int,
        communicator: ZMQCommunicator,
        listener_connection: Tuple[str, str, str],
        *args,
        **kwargs,
    ):
        """Initialize attributes.

        :attr data_id: id of the Data object we are processing.
        :attr communicator: the communicator object for communication with the listener.
        :attr listener_connection: tuple (host, port, protocol) determining the address
            of the listener.
        """
        self.data_id = data_id
        self.process: Dict = PROCESS
        process_requirements: Dict = self.process.get("requirements", {})
        self.requirements: Dict = process_requirements.get("executor", {}).get(
            self.name, {}
        )
        self.resources: Dict = process_requirements.get("resources", {})
        self.storage_url = Path(LOCATION_SUBPATH)
        self.runtime_dir = Path(SETTINGS["FLOW_VOLUMES"]["runtime"]["config"]["path"])
        self.communicator = communicator
        self.listener_connection = list(listener_connection)
        if not sys.platform.startswith("linux"):
            self.listener_connection[0] = "host.docker.internal"

    def _generate_container_name(self, prefix: str) -> str:
        """Generate unique container name."""
        return "{}-{}".format(prefix, self.data_id)

    def get_tools_paths(self):
        """Get tools paths."""
        return SETTINGS["FLOW_EXECUTOR_TOOLS_PATHS"]

    async def start(self):
        """Start process execution."""

    async def run(self):
        """Execute the script and save results."""
        logger.debug("Executor for Data with id %d has started.", self.data_id)
        await self.start()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-10-01 03:15
from __future__ import unicode_literals

from django.db import migrations


def migrate_flow_collection(apps, schema_editor):
    """Migrate 'flow_collection' field to 'entity_type'."""
    Process = apps.get_model("flow", "Process")
    DescriptorSchema = apps.get_model("flow", "DescriptorSchema")

    for process in Process.objects.all():
        process.entity_type = process.flow_collection
        process.entity_descriptor_schema = process.flow_collection

        if (
            process.entity_descriptor_schema is not None
            and not DescriptorSchema.objects.filter(
                slug=process.entity_descriptor_schema
            ).exists()
        ):
            raise LookupError(
                "Descriptow schema '{}' referenced in 'entity_descriptor_schema' not "
                "found.".format(process.entity_descriptor_schema)
            )

        process.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0022_process_entity_1"),
    ]

    operations = [migrations.RunPython(migrate_flow_collection)]



""".. Ignore pydocstyle D400.

=====================================
Django Channels Routing Configuration
=====================================

"""

from channels.routing import ChannelNameRouter, ProtocolTypeRouter

from resolwe.flow.managers import state
from resolwe.flow.managers.consumer import ManagerConsumer

channel_routing = ProtocolTypeRouter(
    {
        "channel": ChannelNameRouter(
            {
                state.MANAGER_CONTROL_CHANNEL: ManagerConsumer,
            }
        ),
    }
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-19 09:37
from __future__ import unicode_literals

from django.db import migrations

from resolwe.flow.models.utils import hydrate_size
from resolwe.flow.utils import iterate_fields


def calculate_total_size(apps, schema_editor):
    """Add ``total_size`` field to all file/dir-type outputs."""
    Data = apps.get_model("flow", "Data")
    for data in Data.objects.all():
        hydrate_size(data, force=True)
        data.save()


def remove_total_size(apps, schema_editor):
    """Remove ``total_size`` field from all file/dir-type outputs."""
    Data = apps.get_model("flow", "Data")
    for data in Data.objects.all():
        for field_schema, fields in iterate_fields(
            data.output, data.process.output_schema
        ):
            name = field_schema["name"]
            value = fields[name]
            if "type" in field_schema:
                if field_schema["type"].startswith("basic:file:"):
                    del value["total_size"]
                elif field_schema["type"].startswith("list:basic:file:"):
                    for obj in value:
                        del obj["total_size"]
                elif field_schema["type"].startswith("basic:dir:"):
                    del value["total_size"]
                elif field_schema["type"].startswith("list:basic:dir:"):
                    for obj in value:
                        del obj["total_size"]
        data.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0005_data_dependency_3"),
    ]

    operations = [
        migrations.RunPython(calculate_total_size, remove_total_size),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-12-06 05:15
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import fernet_fields.fields
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("flow", "0009_make_size_mandatory"),
    ]

    operations = [
        migrations.CreateModel(
            name="Secret",
            fields=[
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "handle",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("value", fernet_fields.fields.EncryptedTextField()),
                (
                    "metadata",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                ("expires", models.DateTimeField(db_index=True, null=True)),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



"""Settings for storage app.

Used to provide s simple default configuration.
"""
from pathlib import Path

from django.conf import settings

from resolwe.flow.executors import constants

project_root = Path(getattr(settings, "PROJECT_ROOT", "/"))
local_connector = "local"
local_dir = project_root / ".test_data"

upload_connector = "upload"
upload_dir = project_root / ".test_upload"

default_storage_connectors = {
    local_connector: {
        "connector": "resolwe.storage.connectors.localconnector.LocalFilesystemConnector",
        "config": {
            "priority": 100,
            "path": local_dir,
            "selinux_label": "z",
        },
    },
    upload_connector: {
        "connector": "resolwe.storage.connectors.localconnector.LocalFilesystemConnector",
        "config": {
            "priority": 100,
            "path": upload_dir,
            "selinux_label": "z",
        },
    },
}

STORAGE_CONNECTORS = getattr(settings, "STORAGE_CONNECTORS", default_storage_connectors)

default_storages = {
    "data": {"connectors": ["local"]},
    "upload": {"connectors": ["upload"]},
}

FLOW_STORAGE = getattr(settings, "FLOW_STORAGE", default_storages)

# This entry must contain the key 'path' inside 'config' dictionary. The value
# is the path where runtime volume is accesible at the host that is executing
# the code (main server for kubernetes, worker nodes for celery/slurm ...).
# Types 'host_path' and 'persistent_volume' are supported.
default_runtime_volume = {
    "type": "host_path",
    "config": {
        "path": project_root / ".test_runtime",
        "name": constants.RUNTIME_VOLUME_NAME,
        "read_only": True,
    },
}

default_processing_volume = {
    "type": "host_path",
    "config": {
        "path": project_root / ".test_processing",
        "name": constants.PROCESSING_VOLUME_NAME,
    },
}

default_secrets_volume = {
    "type": "temporary_directory",
    "config": {"name": constants.SECRETS_VOLUME_NAME, "selinux_label": "z"},
}

default_sockets_volume = {
    "type": "temporary_directory",
    "config": {"name": constants.SOCKETS_VOLUME_NAME, "selinux_label": "z"},
}

default_volumes = {
    constants.RUNTIME_VOLUME_NAME: default_runtime_volume,
    constants.PROCESSING_VOLUME_NAME: default_processing_volume,
    constants.SECRETS_VOLUME_NAME: default_secrets_volume,
    constants.SOCKETS_VOLUME_NAME: default_sockets_volume,
}

FLOW_VOLUMES = getattr(settings, "FLOW_VOLUMES", default_volumes)



# Generated by Django 2.2.10 on 2020-02-24 12:30

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="FileStorage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True, default=django.utils.timezone.now
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="StorageLocation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("url", models.CharField(max_length=60, null=False, blank=False)),
                (
                    "connector_name",
                    models.CharField(blank=False, max_length=30, null=False),
                ),
                (
                    "file_storage",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="storage_locations",
                        to="storage.FileStorage",
                    ),
                ),
                ("last_update", models.DateTimeField(auto_now=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PR", "Preparing"),
                            ("UP", "Uploading"),
                            ("OK", "Done"),
                            ("DE", "Deleting"),
                        ],
                        default="PR",
                        max_length=2,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AccessLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("started", models.DateTimeField(auto_now_add=True)),
                ("finished", models.DateTimeField(blank=True, null=True)),
                ("reason", models.CharField(max_length=120)),
                (
                    "storage_location",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="access_logs",
                        to="storage.StorageLocation",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ReferencedPath",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("path", models.TextField(db_index=True)),
                (
                    "file_storage",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="files",
                        to="storage.FileStorage",
                    ),
                ),
                ("size", models.BigIntegerField(default=-1)),
            ],
        ),
        migrations.AlterField(
            model_name="filestorage",
            name="created",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterUniqueTogether(
            name="storagelocation",
            unique_together={("url", "connector_name")},
        ),
        migrations.AlterModelManagers(
            name="storagelocation",
            managers=[
                ("all_objects", django.db.models.manager.Manager()),
            ],
        ),
    ]



# pylint: disable=missing-docstring
import asyncio
import logging
import os
import subprocess
import sys
import threading
import unittest
from contextlib import suppress
from pathlib import Path
from time import sleep
from unittest import mock

import zmq
import zmq.asyncio
from asgiref.sync import async_to_sync

from django.conf import settings
from django.test import override_settings

from guardian.shortcuts import assign_perm

from resolwe.flow.executors.prepare import BaseFlowExecutorPreparer
from resolwe.flow.executors.socket_utils import Message
from resolwe.flow.executors.startup_processing_container import ProcessingManager
from resolwe.flow.executors.zeromq_utils import ZMQCommunicator
from resolwe.flow.managers import manager
from resolwe.flow.managers.utils import disable_auto_calls
from resolwe.flow.models import Data, DataDependency, Process, Worker
from resolwe.storage import settings as storage_settings
from resolwe.test import (
    ProcessTestCase,
    TestCase,
    tag_process,
    with_docker_executor,
    with_null_executor,
)

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")
DESCRIPTORS_DIR = os.path.join(os.path.dirname(__file__), "descriptors")


class GetToolsTestCase(TestCase):
    @mock.patch("resolwe.flow.utils.apps")
    @mock.patch("resolwe.flow.utils.Path")
    @mock.patch("resolwe.flow.utils.settings")
    def test_get_tools_paths(self, settings_mock, path_mock, apps_mock):
        def path_mock_side_effect(*args):
            if args[0] == "/custom_tools":
                return mock.MagicMock(path=args[0])
            else:
                return app_config_path_mock

        path_mock.side_effect = path_mock_side_effect

        app_config1_mock = mock.MagicMock()
        app_config1_mock.name = "/resolwe/test_app1"
        app_config2_mock = mock.MagicMock()
        app_config2_mock.name = "/resolwe/test_app1"

        apps_mock.get_app_configs.return_value = [app_config1_mock, app_config2_mock]
        first_mock = mock.MagicMock(is_dir=lambda: False, path="/resolwe/test_app1")
        second_mock = mock.MagicMock(is_dir=lambda: True, path="/resolwe/test_app2")

        app_config_path_mock = mock.MagicMock()
        app_config_path_mock.__truediv__.side_effect = [first_mock, second_mock]

        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = ["/custom_tools"]
        base_executor = BaseFlowExecutorPreparer()
        tools_list = [mock.path for mock in base_executor.get_tools_paths()]

        self.assertEqual(len(tools_list), 2)
        self.assertIn("/resolwe/test_app2", tools_list)
        self.assertIn("/custom_tools", tools_list)

    @mock.patch("resolwe.flow.utils.apps")
    @mock.patch("resolwe.flow.utils.settings")
    def test_not_list(self, settings_mock, apps_mock):
        apps_mock.get_app_configs.return_value = []
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = "/custom_tools"

        base_executor = BaseFlowExecutorPreparer()
        with self.assertRaisesRegex(KeyError, "setting must be a list"):
            base_executor.get_tools_paths()


class ManagerRunProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(
            processes_paths=[PROCESSES_DIR], descriptors_paths=[DESCRIPTORS_DIR]
        )

    @with_docker_executor
    @tag_process("test-min-python34")
    def test_python_34(self):
        """Test that processing container starts on Python 3.4.

        This test can be removed when we stop using
        broadinstitute/genomes-in-the-cloud:2.3.1-1504795437 docker image in
        the GATK3 pipeline.
        """
        data = self.run_process("test-min-python34")
        self.assertEqual(data.output["out"], "OK")

    @tag_process("test-min")
    def test_minimal_process(self):
        self.run_process("test-min")

    @tag_process("test-annotate")
    def test_annotate(self):
        data = self.run_process("test-annotate")
        self.assertIsNotNone(data.entity)
        dsc = data.entity.descriptor
        self.assertIn("general", dsc)
        self.assertIn("species", dsc["general"])
        self.assertEqual(dsc["general"]["species"], "Valid")

    def test_router_handover(self):
        """Test router socket handover.

        When multiple clients with the same identity connect to the socket, the
        last one must replace the previous one.
        """
        listener_settings = getattr(settings, "FLOW_EXECUTOR", {}).get(
            "LISTENER_CONNECTION", {}
        )
        port = listener_settings.get("port", 53893)
        host = listener_settings.get("hosts", {}).get("local", "127.0.0.1")
        protocol = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
            "protocol", "tcp"
        )
        logger = logging.getLogger(__name__)
        logger.handlers = []

        async def send_single_message():
            """Open connection to listener and send single message."""
            connection_string = f"{protocol}://{host}:{port}"
            zmq_context = zmq.asyncio.Context.instance()
            zmq_socket = zmq_context.socket(zmq.DEALER)
            zmq_socket.setsockopt(zmq.IDENTITY, b"1")
            zmq_socket.connect(connection_string)
            communicator = ZMQCommunicator(
                zmq_socket, "init_container <-> listener", logger
            )
            async with communicator:
                await asyncio.ensure_future(
                    communicator.send_command(Message.command("update_status", "PP"))
                )

        async def send_two_messages():
            """Send two messages with the same identity.

            When handover is not working the second connection will be rejected
            and no answer received, triggering timeout and raising exception.
            """
            for _ in range(2):
                await send_single_message()

        return asyncio.new_event_loop().run_until_complete(send_two_messages())

    @tag_process("test-annotate-wrong-option")
    def test_annotate_wrong_option(self):
        data = self.run_process(
            "test-annotate-wrong-option", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn("must match one of predefined choices", data.process_error[0])

    @tag_process("test-annotate-wrong-type")
    def test_annotate_wrong_type(self):
        data = self.run_process(
            "test-annotate-wrong-type", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn("is not of type", data.process_error[0])

    @tag_process("test-annotate-missing-field")
    def test_annotate_missing_field(self):
        data = self.run_process(
            "test-annotate-missing-field", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn("definition (invalid) missing in schema", data.process_error[0])

    @tag_process("test-annotate-no-entity")
    def test_annotate_no_entity(self):
        data = self.run_process(
            "test-annotate-no-entity", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn("No entity to annotate", data.process_error[0])

    @tag_process("test-missing-file")
    def test_missing_file(self):
        self.run_process("test-missing-file", assert_status=Data.STATUS_ERROR)

        data = Data.objects.last()
        self.assertEqual(data.status, Data.STATUS_ERROR)

        self.assertEqual(len(data.process_error), 1)
        self.assertIn(
            "Output 'output' set to a missing file: 'i-dont-exist.zip'.",
            data.process_error[0],
        )

    @tag_process("test-spawn-new")
    def test_spawn(self):
        self.run_process("test-spawn-new", tags=["test-tag"])

        data = Data.objects.last()
        file_path = data.location.get_path(filename="foo.bar")
        self.assertEqual(data.output["saved_file"]["file"], "foo.bar")
        self.assertTrue(os.path.isfile(file_path))
        with open(file_path, "rt") as stream:
            self.assertEqual(stream.read(), "foo.bar\n")
        self.assertEqual(data.tags, ["test-tag"])
        parent_data = Data.objects.first()
        self.assertEqual(data.parents.count(), 1)
        self.assertEqual(data.parents.first(), parent_data)

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )

    @tag_process("test-spawn-missing-file")
    def test_spawn_missing_export(self):
        data = self.run_process(
            "test-spawn-missing-file", assert_status=Data.STATUS_ERROR
        )
        self.assertIn(
            "Error while preparing spawned Data objects", data.process_error[0]
        )

    @tag_process(
        "test-broken",
        "test-broken-invalid-execution-engine",
        "test-broken-invalid-expression-engine",
        "test-broken-data-name",
    )
    def test_broken(self):
        Process.objects.create(
            slug="test-broken-invalid-execution-engine",
            name="Test Process",
            contributor=self.contributor,
            type="data:test",
            version=1,
            run={
                "language": "invalid",
            },
        )

        self.run_process("test-broken", assert_status=Data.STATUS_ERROR)
        self.run_process(
            "test-broken-invalid-expression-engine", assert_status=Data.STATUS_ERROR
        )
        self.run_process(
            "test-broken-invalid-execution-engine", assert_status=Data.STATUS_ERROR
        )

        # If evaluation of data_name template fails, the process should not abort as the
        # template may be evaluatable later when the process completes.
        self.run_process("test-broken-data-name")

    @tag_process("test-broken-invalide-storage")
    def test_invalid_storage_file(self):
        data = self.run_process(
            "test-broken-invalide-storage", assert_status=Data.STATUS_ERROR
        )

        self.assertEqual(data.status, Data.STATUS_ERROR)
        print("ERROR", data.process_error[0])
        self.assertIn("must be a valid JSON, current: 1a", data.process_error[0])

    @tag_process("test-workflow-1")
    def test_workflow(self):
        assign_perm("view_collection", self.contributor, self.collection)
        assign_perm("view_collection", self.group, self.collection)

        workflow_data = self.run_process(
            "test-workflow-1", {"param1": "world"}, tags=["test-tag"]
        )

        workflow_data.refresh_from_db()
        step1_data = Data.objects.get(process__slug="test-example-1")
        step2_data = Data.objects.get(process__slug="test-example-2")

        # Workflow should output indices of all data objects, in order.
        self.assertEqual(workflow_data.output["steps"], [step1_data.pk, step2_data.pk])

        # Steps should execute with the correct variables.

        self.assertEqual(step1_data.input["param1"], "world")
        self.assertEqual(step1_data.input["param2"], True)
        self.assertEqual(step1_data.output["out1"], "hello world")
        self.assertEqual(step1_data.tags, ["test-tag"])
        self.assertEqual(step2_data.input["param1"], step1_data.pk)
        self.assertEqual(step2_data.input["param2"]["a"], step1_data.pk)
        self.assertEqual(step2_data.input["param2"]["b"], "hello")
        self.assertEqual(step2_data.output["out1"], "simon says: hello world")
        self.assertEqual(step2_data.tags, ["test-tag"])

        self.assertEqual(step1_data.parents.count(), 1)
        self.assertEqual(step1_data.parents.first(), workflow_data)
        self.assertEqual(step2_data.parents.count(), 2)
        self.assertCountEqual(step2_data.parents.all(), [workflow_data, step1_data])

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in step1_data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )
        self.assertEqual(
            {d.kind for d in step2_data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS, DataDependency.KIND_IO},
        )

        self.assertTrue(self.contributor.has_perm("flow.view_data", step1_data))
        # User inherites permission from group
        self.assertTrue(self.user.has_perm("flow.view_data", step1_data))

    @tag_process("test-workflow-2")
    def test_workflow_entity(self):
        with self.preparation_stage():
            # `self.run_process` adds input data in self.collection
            input_data = self.run_process("test-example-3")

        workflow = self.run_process("test-workflow-2", {"data1": input_data.pk})

        # Check that workflow results are added to the collection.
        self.assertEqual(
            self.collection.data.filter(pk__in=workflow.output["steps"]).count(), 1
        )
        # self.collection now contains workflow and two "normal" data objects
        self.assertEqual(self.collection.data.all().count(), 3)

    @with_docker_executor
    @tag_process("test-docker")
    def test_run_in_docker(self):
        data = self.run_process("test-docker")
        self.assertEqual(data.output["result"], "OK")

    @with_docker_executor
    @disable_auto_calls()
    def test_terminate_worker(self):
        process = Process.objects.get(slug="test-terminate")
        data = Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=process,
        )

        def start_processing(data):
            async_to_sync(manager.communicate)(data_id=data.pk, run_sync=True)

        processing_thread = threading.Thread(target=start_processing, args=(data,))
        processing_thread.start()

        # Wait for up to 5s for process to start.
        for _ in range(50):
            sleep(0.1)
            data.refresh_from_db()
            with suppress(Data.worker.RelatedObjectDoesNotExist):
                if data.worker.status == Worker.STATUS_PROCESSING:
                    break

        self.assertEqual(data.worker.status, Worker.STATUS_PROCESSING)
        data.worker.terminate()

        # Give it max 10 seconds to terminate.
        for _ in range(100):
            sleep(0.1)
            data.refresh_from_db()
            if data.worker.status == Worker.STATUS_COMPLETED:
                break

        self.assertEqual(data.worker.status, Worker.STATUS_COMPLETED)
        self.assertEqual(data.status, Data.STATUS_ERROR)
        self.assertEqual(data.process_error[0], "Processing was cancelled.")
        processing_thread.join(timeout=10)
        self.assertFalse(processing_thread.is_alive())

    @with_docker_executor
    @tag_process("test-requirements-docker")
    def test_executor_requirements(self):
        data = self.run_process("test-requirements-docker")
        self.assertEqual(data.output["result"], "OK")

    @with_docker_executor
    @tag_process("test-docker-uid-gid")
    def test_docker_uid_gid(self):
        data = self.run_process("test-docker-uid-gid")
        self.assertEqual(data.output["result"], "OK")

    @unittest.skip("Null executor test currently not working.")
    @with_null_executor
    @tag_process("test-save-number")
    def test_null_executor(self):
        data = self.run_process(
            "test-save-number", {"number": 19}, assert_status=Data.STATUS_WAITING
        )
        self.assertEqual(data.input["number"], 19)
        self.assertEqual(data.output, {})

    @unittest.skipIf(
        os.environ.get("GITHUB_ACTIONS", "") == "true", "Fails on Github Actions"
    )
    @with_docker_executor
    @tag_process("test-memory-resource-alloc", "test-memory-resource-noalloc")
    def test_memory_resource(self):
        # This process should be terminated due to too much memory usage.
        self.run_process("test-memory-resource-alloc", assert_status=Data.STATUS_ERROR)
        # This process should run normally (honors the limits).
        self.run_process("test-memory-resource-noalloc")

    @with_docker_executor
    @tag_process("test-cpu-resource-1core", "test-cpu-resource-2core")
    def test_cpu_resource(self):
        # Currently there is no good way to test this reliably, so we just check if the
        # resource limit specification still makes the process run.
        self.run_process("test-cpu-resource-1core")
        self.run_process("test-cpu-resource-2core")

    @unittest.skipIf(
        settings.FLOW_DOCKER_DISABLE_SECCOMP, "Docker seccomp is disabled."
    )
    @with_docker_executor
    @tag_process(
        "test-network-resource-enabled",
        "test-network-resource-disabled",
        "test-network-resource-policy",
    )
    def test_network_resource(self):
        self.run_process("test-network-resource-enabled")
        self.run_process(
            "test-network-resource-disabled", assert_status=Data.STATUS_ERROR
        )
        self.run_process(
            "test-network-resource-policy", assert_status=Data.STATUS_ERROR
        )

    @with_docker_executor
    @override_settings(FLOW_PROCESS_RESOURCE_DEFAULTS={"cpu_time_interactive": 1})
    @tag_process(
        "test-scheduling-class-interactive-ok",
        "test-scheduling-class-interactive-fail",
        "test-scheduling-class-batch",
    )
    def test_scheduling_class(self):
        self.run_process("test-scheduling-class-interactive-ok")
        self.run_process(
            "test-scheduling-class-interactive-fail", assert_status=Data.STATUS_ERROR
        )
        self.run_process("test-scheduling-class-batch")

    @with_docker_executor
    @tag_process("test-save-number")
    def test_executor_fs_lock(self):
        # First, run the process normaly. Do not autoremove container to check
        # if name clash with init contaner is avoided.
        with self.settings(FLOW_DOCKER_AUTOREMOVE=False):
            data = self.run_process("test-save-number", {"number": 42})
        # Make sure that process was successfully ran first time.
        self.assertEqual(data.output["number"], 42)
        data.output = {}
        data.save()

        file = Path(data.location.get_path(filename="temporary_file_do_not_purge.txt"))
        self.assertFalse(file.exists())
        file.touch()

        listener_settings = getattr(settings, "FLOW_EXECUTOR", {}).get(
            "LISTENER_CONNECTION", {}
        )
        port = listener_settings.get("port", 53893)
        hosts = listener_settings.get("hosts", {"local": "127.0.0.1"})
        host = hosts.get("local", next(iter(hosts.values())))
        protocol = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
            "protocol", "tcp"
        )
        process = subprocess.run(
            [
                "python",
                "-m",
                "executors",
                ".docker",
                str(data.id),
                host,
                str(port),
                protocol,
            ],
            cwd=storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=30,
        )
        self.assertEqual(process.returncode, 0)
        data.refresh_from_db()
        self.assertEqual(data.output, {})
        self.assertEqual(data.status, Data.STATUS_DONE)
        self.assertEqual(data.process_error, [])
        self.assertEqual(data.process_info, [])
        # Check that temporary file was not deleted.
        self.assertTrue(file.exists())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_communication_connection_ok(self, open_mock):
        """Test connection under normal conditions."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            await manager._wait_for_communication_container(None)

        async def readline():
            """Return PING."""
            return b"PING"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline), None)

        reader_mock = mock.MagicMock()
        reader_mock.readline = readline
        if sys.version_info >= (3, 8):
            open_mock.return_value = (mock.Mock(readline=readline), None)
        else:
            open_mock.return_value = reader_writer()
        loop = asyncio.new_event_loop()
        loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_communication_connection_ok_script(self, open_mock):
        """Test connection if first attempt fails."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            await manager._wait_for_communication_container(None)

        async def readline_ok():
            """Return PING."""
            return b"PING"

        async def readline_error():
            """Return ERROR."""
            return b"ERROR"

        async def reader_writer(coroutine):
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=coroutine), None)

        reader_mock = mock.MagicMock()
        reader_mock.readline.side_effect = [readline_error(), readline_ok()]
        if sys.version_info >= (3, 8):
            open_mock.return_value = (reader_mock, None)
        else:
            open_mock.side_effect = [
                reader_writer(readline_error),
                reader_writer(readline_ok),
            ]
        loop = asyncio.new_event_loop()
        loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_communication_connection_fail(self, open_mock):
        """Test function behaviour when connection fails."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            await manager._wait_for_communication_container(None)

        async def readline_error():
            """Return ERROR."""
            return b"ERROR"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline_error), None)

        reader_mock = mock.MagicMock()
        reader_mock.readline.side_effect = [readline_error(), readline_error()]
        if sys.version_info >= (3, 8):
            open_mock.return_value = (reader_mock, None)
        else:
            open_mock.side_effect = [reader_writer(), reader_writer()]
        loop = asyncio.new_event_loop()
        with self.assertRaisesRegex(RuntimeError, "Communication .* unreacheable"):
            loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_communication_connection_fail_script(self, open_mock):
        """Test script behaviour if connection fails."""

        async def main_test():
            """Main test coroutine."""
            protocol_mock = mock.Mock(terminate_script=terminate_script)
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            manager.protocol_handler = protocol_mock
            await manager.run()

        async def terminate_script():
            """Terminate script mock."""

        async def readline_error():
            """Return ERROR."""
            return b"ERROR"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline_error), None)

        reader_mock = mock.MagicMock()
        reader_mock.readline.side_effect = [readline_error(), readline_error()]
        if sys.version_info >= (3, 8):
            open_mock.return_value = (reader_mock, None)
        else:
            open_mock.side_effect = [reader_writer(), reader_writer()]
        loop = asyncio.new_event_loop()
        with self.assertRaisesRegex(RuntimeError, "Communication .* unreacheable"):
            loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch("resolwe.flow.executors.startup_processing_container.socket")
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_upload_fail(self, socket_mock, open_mock):
        """Test startup script if opening upload socket fails."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            await manager.run()

        async def readline():
            """Return PING."""
            return b"PING"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline), None)

        socket_mock.socket.return_value.connect.side_effect = RuntimeError
        if sys.version_info >= (3, 8):
            open_mock.return_value = (mock.Mock(readline=readline), None)
        else:
            open_mock.return_value = reader_writer()
        loop = asyncio.new_event_loop()
        with self.assertRaises(RuntimeError):
            loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch("resolwe.flow.executors.startup_processing_container.socket")
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.start_unix_server"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_script_connection_fail(
        self, server_mock, socket_mock, open_mock
    ):
        """Test startup script if opening processing script socket fails."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            await manager.run()

        async def readline():
            """Return PING."""
            return b"PING"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline), None)

        socket_mock.socket.return_value.connect.side_effect = None
        if sys.version_info >= (3, 8):
            open_mock.return_value = (mock.Mock(readline=readline), None)
        else:
            open_mock.return_value = reader_writer()
        loop = asyncio.new_event_loop()
        server_mock.side_effect = ValueError("test")
        with self.assertRaisesRegex(ValueError, "test"):
            loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch("resolwe.flow.executors.startup_processing_container.socket")
    @mock.patch("resolwe.flow.executors.startup_processing_container.create_task")
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.start_unix_server"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_create_task_called(
        self, server_mock, create_mock, socket_mock, open_mock
    ):
        """Test that upload timer task is created."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            await manager.run()

        async def readline():
            """Return PING."""
            return b"PING"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline), None)

        socket_mock.socket.return_value.connect.side_effect = None
        if sys.version_info >= (3, 8):
            open_mock.return_value = (mock.Mock(readline=readline), None)
        else:
            open_mock.return_value = reader_writer()
        loop = asyncio.new_event_loop()
        create_mock.side_effect = ValueError("test")
        with self.assertRaisesRegex(ValueError, "test"):
            loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch("resolwe.flow.executors.startup_processing_container.socket")
    @mock.patch("resolwe.flow.executors.startup_processing_container.create_task")
    @mock.patch("resolwe.flow.executors.startup_processing_container.ProtocolHandler")
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.start_unix_server"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_communicate(
        self, server_mock, protocol_mock, create_mock, socket_mock, open_mock
    ):
        """Test that communicate is called."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            manager = ProcessingManager(current_loop)
            await manager.run()

        async def readline():
            """Return PING."""
            return b"PING"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline), None)

        socket_mock.socket.return_value.connect.side_effect = None
        if sys.version_info >= (3, 8):
            open_mock.return_value = (mock.Mock(readline=readline), None)
        else:
            open_mock.return_value = reader_writer()
        protocol_mock.return_value.communicate.side_effect = ValueError("test")
        loop = asyncio.new_event_loop()
        with self.assertRaisesRegex(ValueError, "test"):
            loop.run_until_complete(main_test())



"""Processing container startup script."""
import array
import asyncio
import concurrent
import functools
import json
import logging
import os
import shlex
import shutil
import signal
import socket
import sys
import time
import uuid
from contextlib import suppress
from distutils.util import strtobool
from pathlib import Path

try:
    import constants
except ImportError:
    from resolwe.flow.executors import constants

# The directory where local processing is done.
TMP_DIR = Path(os.environ.get("TMP_DIR", ".tmp"))
DATA_ID = int(os.getenv("DATA_ID", "-1"))

# Basic constants.
STDOUT_LOG_PATH = constants.PROCESSING_VOLUME / "stdout.txt"
JSON_LOG_PATH = constants.PROCESSING_VOLUME / "jsonout.txt"

# Update log files every 30 seconds.
UPDATE_LOG_FILES_TIMEOUT = 30

# Keep data settings.
KEEP_DATA = bool(strtobool(os.environ.get("FLOW_MANAGER_KEEP_DATA", "False")))

# Read configuration from environmental variables.
COMMUNICATION_SOCKET = (
    constants.SOCKETS_VOLUME / constants.COMMUNICATION_PROCESSING_SOCKET
)
SCRIPT_SOCKET = constants.SOCKETS_VOLUME / constants.SCRIPT_SOCKET
UPLOAD_FILE_SOCKET = constants.SOCKETS_VOLUME / constants.UPLOAD_FILE_SOCKET

# How many file descriptors to sent over socket in a single message.
DESCRIPTOR_CHUNK_SIZE = int(os.environ.get("DESCRIPTOR_CHUNK_SIZE", 100))


logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger(__name__)


assert sys.version_info[0] == 3, "Only Python3 is supported"
assert sys.version_info[1] >= 4, "Only Python >= 3.4 is supported"

if sys.version_info[1] == 4:
    # This is done to awoid syntax error in Python >= 3.7
    create_task = getattr(asyncio, "async")
else:
    create_task = asyncio.ensure_future


def respond(message, status, response_data):
    """Return a response to the given message."""
    if message["type"] in ["COMMAND", "EQR"]:
        response_type = "RESPONSE"
    else:
        response_type = "HBTR"
    response = {"type": response_type}
    response["type_data"] = status
    response["data"] = response_data
    response["uuid"] = message.get("uuid", uuid.uuid4().hex)
    if "sequence_number" in message:
        response["sequence_number"] = message["sequence_number"]
    response["timestamp"] = time.time()
    return response


class ProtocolHandler:
    """Handles communication with the communication container."""

    def __init__(self, manager):
        # type: (ProcessingManager) -> None
        """Initialize."""
        self._script_future = None  # type: Optional[asyncio.Future] # noqa: F821
        self.return_code = 0
        self.manager = manager
        self.communicator = manager.communicator
        self._script_finishing = None
        self._response_queue = list()
        self._command_lock = asyncio.Lock()
        self._message_in_queue = asyncio.Event()
        self._command_counter = 0

    @asyncio.coroutine
    def communicate(self):
        """Listen for messages from communication container."""
        while True:
            message = yield from self.communicator.receive_data()
            if message is None:
                break
            if message["type"] == "HBT":
                logger.debug("Responding to HBT.")
                yield from self.communicator.send_data(respond(message, "OK", ""))
            elif message["type"] == "EQR":
                logger.debug("Responding to EQR.")
                yield from self.communicator.send_data(
                    respond(message, "OK", "received")
                )
            elif message["type"] == "COMMAND":
                command_name = message["type_data"]
                if command_name == "process_script":
                    if self._script_future is not None:
                        response = respond(message, "ERR", "Already processing")
                        yield from self.communicator.send_data(response)
                    else:
                        create_task(self.process_script(message))
                elif command_name == "terminate":
                    logger.debug("Got terminate command, shutting down.")
                    self._command_counter += 1
                    yield from self.terminate_script(
                        {"info": "Processing was cancelled."}
                    )
                    response = respond(message, "OK", "")
                    yield from self.communicator.send_data(response)
                    # Wait for script to do the clean-up.
                    yield from self._script_finishing.wait()
                    break
                else:
                    logger.error("Unknown command: %s.", command_name)
                    response = respond(message, "ERR", "Unknown command")
                    yield from self.communicator.send_data(response)
                    break
            elif message["type"] == "RESPONSE":
                self._response_queue.append(message)
                self._message_in_queue.set()

        logger.debug("Communication with communication container stopped.")

    @asyncio.coroutine
    def send_command(self, command):
        """Send a command and return a response."""
        logger.debug("Sending command %s.", command)
        with (yield from self._command_lock):
            self._command_counter += 1
            command["uuid"] = uuid.uuid4().hex
            command["sequence_number"] = self._command_counter
            command["timestamp"] = time.time()
            try:
                assert len(self._response_queue) == 0
                self._message_in_queue.clear()
                yield from self.communicator.send_data(command)
                yield from self._message_in_queue.wait()
                return self._response_queue.pop()
            except:
                logger.exception("Unknown exception in send_command.")
                raise

    @asyncio.coroutine
    def _send_log(self, message):
        """Send log message."""
        log_command = {
            "type": "COMMAND",
            "type_data": "process_log",
            "data": message,
        }
        yield from self.send_command(log_command)

    @asyncio.coroutine
    def process_script(self, message):
        # type: (dict) -> ()
        """Process the given script."""
        self.return_code = 1
        self._script_future = create_task(self._execute_script(message["data"]))
        try:
            yield from self._script_future
            self.return_code = self._script_future.result()
        finally:
            self._script_future = None
            try:
                # Cancel sending logs on timer.
                logger.debug("Stopping upload log future.")
                self.manager.update_logs_future.cancel()
                yield from self.manager.update_logs_future
                logger.debug("Uploading latest log files.")
                try:
                    yield from self.manager.upload_log_files()
                except:
                    yield from self._send_log(
                        {"error": "Error uploading final log files."}
                    )
                    logger.exception("Error uploading final log files.")
                    if self.return_code == 0:
                        self.return_code = -1

                if KEEP_DATA:
                    try:
                        yield from self.manager.send_file_descriptors(
                            [
                                str(file)
                                for file in Path("./").rglob("*")
                                if file.is_file()
                            ]
                        )
                    except:
                        logger.exception("Error sending file descriptors (keep_data).")
                        yield from self._send_log(
                            {
                                "error": "Error uploading final file descriptors (keep_data)."
                            }
                        )
                        if self.return_code == 0:
                            self.return_code = -1

                response = respond(message, "OK", self.return_code)
                yield from self.communicator.send_data(response)

            except:
                logger.exception(
                    "Unexpected exception in process_script finally clause."
                )
            finally:
                # Close the cliet socket to notify communicator that processing is
                # finished.
                logger.debug("Closing upload socket.")
                self.manager.upload_socket.close()
                logger.debug("Upload socket closed.")
                if self._script_finishing:
                    logger.debug("Setting script finishing.")
                    self._script_finishing.set()

    @asyncio.coroutine
    def terminate_script(self, message=None):
        """Terminate the running script and optionally send log message."""
        try:
            if message is not None:
                yield from self._send_log(message)
        except:
            logger.exception("Error sending terminating message to listener.")
        finally:
            if self._script_future is not None:
                logger.debug("Terminating script")
                self._script_finishing = asyncio.Event()
                self._script_future.cancel()

    @asyncio.coroutine
    def _execute_script(self, script):
        # type: (str) -> int
        """Execute given bash script.

        :returns: the return code of the script.
        """
        # Start the bash subprocess and write script to its standard input.
        logger.debug("Executing script: %s.", script)
        bash_command = "/bin/bash --login" + os.linesep

        # The following comment commands black to not reformat the code
        # block. It must not have trailing comma due compatibility with
        # Python 3.4.

        # fmt: off
        proc = yield from asyncio.subprocess.create_subprocess_exec(
            *shlex.split(bash_command),
            limit=4 * (2 ** 20),  # 4MB buffer size for line buffering
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT
        )
        # fmt: on
        assert proc.stdin is not None
        assert proc.stdout is not None
        proc.stdin.write(script.encode("utf-8"))
        proc.stdin.close()
        logger.debug("Script sent to stdin.")

        try:
            received = yield from proc.stdout.readline()
            while received:
                with (yield from self.manager._uploading_log_files_lock):
                    with STDOUT_LOG_PATH.open("ab") as stdout_file:
                        stdout_file.write(received)
                        self.manager.log_files_need_upload = True
                received = yield from proc.stdout.readline()
        except asyncio.CancelledError:
            # Task was cancelled: terminate the running subprocess immediately.
            proc.kill()
        finally:
            yield from proc.wait()
            assert proc.returncode is not None
            logger.debug("Script finished with rc %d.", proc.returncode)
            return proc.returncode


class Communicator:
    """Simple communication class."""

    def __init__(self, reader, writer):
        """Initialize."""
        self.reader = reader
        self.writer = writer

    @asyncio.coroutine
    def send_data(self, data, size_bytes=8):
        # type: (dict, int) -> None
        """Send data over socket.

        :param s: socket to send over.
        :param data: dict that must be serialzable to JSON.
        :param size_bytes: how first many bytes in message are dedicated to the
            message size (send as binary number).
        :raises: exception on failure.
        """
        payload = json.dumps(data).encode("utf-8")
        size = len(payload).to_bytes(size_bytes, byteorder="big")
        self.writer.write(size)
        self.writer.write(payload)
        yield from self.writer.drain()

    @asyncio.coroutine
    def receive_data(self, size_bytes=8):
        # type: (socket.SocketType, int) -> Optional[Dict] # noqa: F821
        """Recieve data over the given socket.

        :param size_bytes: how first many bytes in message are dedicated to the
            message size (binary number).
        """
        with suppress(asyncio.IncompleteReadError):
            received = yield from self.reader.readexactly(size_bytes)
            message_size = int.from_bytes(received, byteorder="big")
            received = yield from self.reader.readexactly(message_size)
            assert len(received) == message_size
            return json.loads(received.decode("utf-8"))


class ProcessingManager:
    """Manager class.

    It starts network servers and executes the received script.
    """

    def __init__(self, loop):
        """Initialize."""
        self.worker_reader = None  # type: Optional[asyncio.StreamReader] # noqa: F821
        self.worker_writer = None  # type: Optional[asyncio.StreamWriter] # noqa: F821
        self.protocol_handler = None  # type: Optional[BaseProtocol] # noqa: F821
        self.worker_connected = asyncio.Event()
        self.exported_files_mapper = dict()  # type: Dict[str, str] # noqa: F821
        self.upload_socket = None
        self.update_logs_future = None
        self._uploading_log_files_lock = asyncio.Lock(loop=loop)
        self._send_file_descriptors_lock = asyncio.Lock(loop=loop)
        self.log_files_need_upload = False
        self.loop = loop

    @asyncio.coroutine
    def _wait_for_communication_container(self, path):
        # type: (str) -> Communicator
        r"""Wait for communication container to start.

        Assumption is that the communication container is accepting connections
        and will send a single b"PING\n" when connected.

        :raises RuntimeError: if unable to connect to the communication
            container for COMMUNICATOR_WAIT_TIMEOUT seconds.
        """
        for retry in range(constants.CONTAINER_TIMEOUT):
            try:
                reader, writer = yield from asyncio.open_unix_connection(path)
                line = (yield from reader.readline()).decode("utf-8")
                assert line.strip() == "PING", "Expected 'PING', got '{}'.".format(line)
                return Communicator(reader, writer)
            except:
                # Raise RuntimeError on final retry.
                if retry + 1 == constants.CONTAINER_TIMEOUT:
                    raise RuntimeError("Communication container is unreacheable.")
                with suppress(Exception):
                    writer.close()
                    yield from writer.wait_closed()
                time.sleep(1)

    @asyncio.coroutine
    def run(self):
        # type: () -> int
        """Start the main method.

        :raises RuntimeError: in case of failure.
        """
        try:
            logger.debug("Connecting to the communication container.")
            self.communicator = yield from self._wait_for_communication_container(
                str(COMMUNICATION_SOCKET)
            )
            logger.info("Connected to the communication container.")

            self.protocol_handler = ProtocolHandler(self)

            # Connect to the upload socket.
            self.upload_socket = socket.socket(family=socket.AF_UNIX)
            logger.debug("Connecting upload socket to: %s.", UPLOAD_FILE_SOCKET)
            self.upload_socket.connect(str(UPLOAD_FILE_SOCKET))

            # Accept network connections from the processing script.
            yield from asyncio.start_unix_server(
                self._handle_processing_script_connection,
                str(SCRIPT_SOCKET),
            )

            # Update log files.
            self.update_logs_future = create_task(self.update_log_files_timer())

            # Await messages from the communication controler. The command
            # stops when socket is closed or communication is stopped.
            yield from self.protocol_handler.communicate()
        finally:
            # Just in case connection was closed before script was finished.
            yield from self.protocol_handler.terminate_script()
            logger.debug("Stopping processing container.")
        return self.protocol_handler.return_code

    @asyncio.coroutine
    def _handle_export_files(self, file_names):
        # type: (dict) -> None
        """Handle export files."""
        unique_names = []
        for file_name in file_names:
            unique_name = "export_" + uuid.uuid4().hex
            shutil.move(file_name, unique_name)
            unique_names.append(unique_name)
        presigned_urls = yield from self.send_file_descriptors(
            unique_names, need_presigned_urls=True, storage_name="upload"
        )
        for file_name, presigned_url in zip(file_names, presigned_urls):
            self.exported_files_mapper[file_name] = presigned_url

    @asyncio.coroutine
    def update_log_files_timer(self):
        """Update log files every UPDATE_LOG_FILES_TIMEOUT secods."""
        while True:
            try:
                logger.debug("Uploading log files on timer.")
                yield from self.upload_log_files()
                yield from asyncio.sleep(UPDATE_LOG_FILES_TIMEOUT)
            except asyncio.CancelledError:
                logger.info("Update log_files_timer canceled.")
                break
            except:
                logger.exception("Error uploading log files.")
                yield from self.protocol_handler.terminate_script(
                    {"error": "Error uploading log files."}
                )

    @asyncio.coroutine
    def upload_log_files(self):
        """Upload log files if needed.

        :raises BrokenPipeError: when upload socket is not available.
        """
        if self.log_files_need_upload:
            log_paths = [
                log_file.relative_to(constants.PROCESSING_VOLUME)
                for log_file in [STDOUT_LOG_PATH, JSON_LOG_PATH]
            ]
            logger.debug("Uploading log files %s.", log_paths)
            with (yield from self._uploading_log_files_lock):
                yield from self.send_file_descriptors(log_paths)
            self.log_files_need_upload = False
        else:
            logger.debug("Log files have not changed.")

    @asyncio.coroutine
    def send_file_descriptors(
        self, filenames, need_presigned_urls=False, storage_name="data"
    ):
        # type: (Tuple(str, Union[List[str], PurePath], bool)) -> List  # noqa: F821
        """Send file descriptors over UNIX sockets.

        Code is based on the sample in manual
        https://docs.python.org/3/library/socket.html#socket.socket.sendmsg .

        :returns: the list of presigned urls for filenames (in the same order)
            if need_presigned_urls is set to True. Otherwise empty list is
            returned.

        :raises RuntimeError: on failure.
        """

        def send_chunk(processing_filenames):
            """Send chunk of filenames to uploader.

            :raises RuntimeError: on failure.
            :returns: the list of presigned urls (if requested) or empty list.
            """
            logger.debug("Sending file descriptors for files: %s", processing_filenames)
            message = [storage_name, processing_filenames, need_presigned_urls]
            payload = json.dumps(message).encode()
            # File handlers must be created before file descriptors otherwise
            # garbage collector will kick in and close the handlers. Handlers
            # will be closed when function completes.
            file_handlers = [open(filename, "rb") for filename in processing_filenames]
            file_descriptors = [file_handler.fileno() for file_handler in file_handlers]
            self.upload_socket.sendall(len(payload).to_bytes(8, byteorder="big"))
            self.upload_socket.sendmsg(
                [payload],
                [
                    (
                        socket.SOL_SOCKET,
                        socket.SCM_RIGHTS,
                        array.array("i", file_descriptors),
                    )
                ],
            )
            response_length = int.from_bytes(
                self.upload_socket.recv(8), byteorder="big"
            )
            response = json.loads(self.upload_socket.recv(response_length).decode())
            if not response["success"]:
                raise RuntimeError(
                    "Communication container response indicates error sending "
                    "files {}.".format(processing_filenames)
                )
            return response["presigned_urls"]

        result = []
        with (yield from self._send_file_descriptors_lock):
            with concurrent.futures.ThreadPoolExecutor(max_workers=1) as pool:
                for i in range(0, len(filenames), DESCRIPTOR_CHUNK_SIZE):
                    processing_filenames = [
                        str(file_) for file_ in filenames[i : i + DESCRIPTOR_CHUNK_SIZE]
                    ]
                    response = yield from self.loop.run_in_executor(
                        pool, send_chunk, processing_filenames
                    )
                    if need_presigned_urls:
                        result += response
        return result

    @asyncio.coroutine
    def _handle_processing_script_connection(self, reader, writer):
        # type: (asyncio.StreamReader, asyncio.StreamWriter) -> None
        """Handle incoming connection from the processing script.

        The requests are just proxied to the communication container with
        the exception of the .

        Python process starts a single connection while Resolwe runtime utils
        starts a new connection for every request.
        """
        try:
            communicator = Communicator(reader, writer)
            logger.debug("Processing script connected to socket.")

            message = yield from communicator.receive_data()
            while message:
                # Lock uploading log files for the duration of the processing
                # of the received command.
                # Doing so we avoid issues with processing container issuing
                # two commands at (almost) the same time. Since listener can
                # process received commands in parallel and each of them is
                # executed inside transaction the final state of the data
                # object depends on the exact timing of these transactions.
                # One possible outcome is that the transaction that updates
                # log files is executed while transaction that updates outputs
                # is in progress and finishes last, effectively canceling the
                # first transaction leading to a data loss.

                with (yield from self._uploading_log_files_lock):
                    with JSON_LOG_PATH.open("at") as json_file:
                        json_file.write(json.dumps(message) + os.linesep)
                        self.log_files_need_upload = True
                    logger.debug("Processing script sent message: %s.", message)
                    command = message["type_data"]
                    if command == "export_files":
                        yield from self._handle_export_files(message["data"])
                        response = respond(message, "OK", "")
                    elif command == "upload_files":
                        # Connect to the upload socket.
                        filenames = message["data"]
                        logger.debug("Sending filedescriptors.")
                        # This can block, run in a thread maybe?
                        logger.debug("Names: %s", filenames)
                        yield from self.send_file_descriptors(filenames)
                        logger.debug("File descriptors sent.")
                        response = respond(message, "OK", "")
                    elif command == "run":
                        message["data"] = {
                            "data": message["data"],
                            "export_files_mapper": self.exported_files_mapper,
                        }
                        response = yield from self.protocol_handler.send_command(
                            message
                        )
                    else:
                        response = yield from self.protocol_handler.send_command(
                            message
                        )

                # Terminate the script on ERROR response. This has to be
                # outside the lock block above since it sends latest log files
                # to the listener.
                if response["type_data"] == "ERR":
                    logger.debug("Response with error status received, terminating")
                    # Terminate the script and wait for termination.
                    # Otherwise script may continue to run and produce
                    # hard to debug error messages.
                    yield from self.protocol_handler.terminate_script(
                        {"info": "Response with status 'ERROR' received from listener."}
                    )
                    if self.protocol_handler._script_finishing is not None:
                        yield from self.protocol_handler._script_finishing.wait()
                        break
                else:
                    yield from communicator.send_data(response)
                message = yield from communicator.receive_data()

        except (asyncio.IncompleteReadError, GeneratorExit):
            logger.info("Processing script closed connection.")
        except:
            logger.exception(
                "Unknown exception while handling processing script message."
            )
            yield from self.protocol_handler.terminate_script(
                {
                    "error": "Unexpected exception while handling processing script message."
                }
            )
            yield from self.protocol_handler._script_finishing.wait()
        finally:
            writer.close()


@asyncio.coroutine
def sig_term_handler(processing_manager):
    # type: (asyncio.Future) -> None
    """Gracefully terminate the running process."""
    logger.debug("SIG_INT received, shutting down.")
    yield from processing_manager.protocol_handler.terminate_script(
        {"error": "SIG_INT received, terminating."}
    )


@asyncio.coroutine
def start_processing_container(loop):
    # type: () -> None
    """Start the processing manager and set SIGINT handler."""
    processing_manager = ProcessingManager(loop)
    manager_future = create_task(processing_manager.run())

    asyncio.get_event_loop().add_signal_handler(
        signal.SIGINT,
        functools.partial(create_task, sig_term_handler(processing_manager)),
    )

    # Wait for the manager task to finish.
    return (yield from manager_future)


if __name__ == "__main__":
    # Change working directory into /data_local and create .tmp directory
    # inside it.
    tmp_path = constants.PROCESSING_VOLUME / TMP_DIR
    if not tmp_path.is_dir():
        (constants.PROCESSING_VOLUME / TMP_DIR).mkdir()
    os.chdir(str(constants.PROCESSING_VOLUME))

    # Create log files.
    STDOUT_LOG_PATH.touch()
    JSON_LOG_PATH.touch()

    # Replace the code bellow with asyncio.run when we no longer have to
    # support python 3.6.
    loop = asyncio.get_event_loop()
    return_code = loop.run_until_complete(start_processing_container(loop))
    loop.close()
    sys.exit(return_code)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-03-01 02:25
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0030_add_data_duplicate"),
    ]

    operations = [
        migrations.AddField(
            model_name="entity",
            name="duplicated",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



"""Permissions functions used in Resolwe Viewsets."""
from distutils.util import strtobool

from django.contrib.contenttypes.models import ContentType
from django.db import transaction

from guardian.models import UserObjectPermission
from rest_framework import exceptions, serializers, status
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.permissions.shortcuts import get_object_perms

from .utils import (
    check_owner_permission,
    check_public_permissions,
    check_user_permissions,
    update_permission,
)


class CurrentUserPermissionsSerializer(serializers.Serializer):
    """Current user permissions serializer."""

    id = serializers.IntegerField()
    type = serializers.CharField(max_length=50)
    name = serializers.CharField(max_length=100)
    permissions = serializers.ListField(child=serializers.CharField(max_length=30))


class ResolwePermissionsMixin:
    """Mixin to support managing `Resolwe` objects' permissions."""

    def get_serializer_class(self):
        """Augment base serializer class.

        Include permissions information with objects.

        """
        base_class = super().get_serializer_class()

        class SerializerWithPermissions(base_class):
            """Augment serializer class."""

            def get_fields(serializer_self):
                """Return serializer's fields."""
                fields = super().get_fields()
                fields["current_user_permissions"] = CurrentUserPermissionsSerializer(
                    read_only=True
                )
                return fields

            def to_representation(serializer_self, instance):
                """Object serializer."""
                data = super().to_representation(instance)

                if (
                    "fields" not in self.request.query_params
                    or "current_user_permissions" in self.request.query_params["fields"]
                ):
                    data["current_user_permissions"] = get_object_perms(
                        instance, self.request.user
                    )

                return data

        return SerializerWithPermissions

    def set_content_permissions(self, user, obj, payload):
        """Overwritte this function in sub-classes if needed.

        It will be called if ``share_content`` query parameter will be
        passed with the request. So it should take care of sharing
        content of current object, i.e. data objects and samples
        attached to a collection.
        """

    @action(
        detail=True,
        methods=["get", "post"],
        url_path="permissions",
        url_name="permissions",
    )
    def detail_permissions(self, request, pk=None):
        """Get or set permissions API endpoint."""
        obj = self.get_object()

        if request.method == "POST":
            content_type = ContentType.objects.get_for_model(obj)
            payload = request.data
            share_content = strtobool(payload.pop("share_content", "false"))
            user = request.user
            is_owner = user.has_perm("owner_{}".format(content_type.name), obj=obj)

            allow_owner = is_owner or user.is_superuser
            check_owner_permission(payload, allow_owner)
            check_public_permissions(payload)
            check_user_permissions(payload, request.user.pk)

            with transaction.atomic():
                update_permission(obj, payload)

                owner_count = UserObjectPermission.objects.filter(
                    object_pk=obj.id,
                    content_type=content_type,
                    permission__codename__startswith="owner_",
                ).count()

                if not owner_count:
                    raise exceptions.ParseError("Object must have at least one owner.")

            if share_content:
                self.set_content_permissions(user, obj, payload)

        return Response(get_object_perms(obj))

    @action(
        detail=False,
        methods=["get", "post"],
        url_path="permissions",
        url_name="permissions",
    )
    def list_permissions(self, request):
        """Batch get or set permissions API endpoint."""
        # TODO: Implement batch get/set permissions
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)



"""Process descriptor."""
import collections
import re

# Processor type validation regexp.
PROCESSOR_TYPE_RE = re.compile(r"^data:[a-z0-9:]+$")


class SchedulingClass:
    """Process scheduling class."""

    INTERACTIVE = "interactive"
    BATCH = "batch"


class Persistence:
    """Process data persistance."""

    RAW = "RAW"
    CACHED = "CACHED"
    TEMP = "TEMP"


class ValidationError(Exception):
    """Process descriptor validation error."""


class ProcessDescriptor:
    """Process descriptor.

    The process descriptor is a result of parsing a process definition. It contains
    everything that is required for registering the process into the database.
    """

    class Metadata:
        """General process metadata."""

        slug = None
        name = None
        process_type = None
        description = None
        version = None
        category = None
        scheduling_class = None
        persistence = None
        requirements = None
        data_name = None
        entity = None
        lineno = None

        def __repr__(self):
            """Return string representation."""
            return repr(self.__dict__)

    def __init__(self, source=None, parent_process=None):
        """Construct a new process descriptor."""
        self.metadata = ProcessDescriptor.Metadata()
        self.inputs = collections.OrderedDict()
        self.outputs = collections.OrderedDict()
        self.relations = None
        self.name = None
        self.source = source
        self.parent_process = parent_process

    def validate(self):
        """Validate process descriptor."""
        required_fields = ("slug", "name", "process_type", "version")
        for field in required_fields:
            if getattr(self.metadata, field, None) is None:
                raise ValidationError(
                    "process '{}' is missing required meta attribute: {}".format(
                        self.metadata.slug or "<unknown>", field
                    )
                )

        if not PROCESSOR_TYPE_RE.match(self.metadata.process_type):
            raise ValidationError(
                "process '{}' has invalid type: {}".format(
                    self.metadata.slug, self.metadata.process_type
                )
            )

    def to_schema(self):
        """Return process schema for this process."""
        process_type = self.metadata.process_type
        if not process_type.endswith(":"):
            process_type = "{}:".format(process_type)

        schema = {
            "slug": self.metadata.slug,
            "name": self.metadata.name,
            "type": process_type,
            "version": self.metadata.version,
            "data_name": "",
            "requirements": {
                "executor": {
                    "docker": {
                        "image": "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-18.04",
                    },
                },
            },
        }

        if self.metadata.description is not None:
            schema["description"] = self.metadata.description
        if self.metadata.category is not None:
            schema["category"] = self.metadata.category
        if self.metadata.scheduling_class is not None:
            schema["scheduling_class"] = self.metadata.scheduling_class
        if self.metadata.persistence is not None:
            schema["persistence"] = self.metadata.persistence
        if self.metadata.requirements is not None:
            schema["requirements"] = self.metadata.requirements
        if self.metadata.data_name is not None:
            schema["data_name"] = self.metadata.data_name
        if self.metadata.entity is not None:
            schema["entity"] = self.metadata.entity

        if self.inputs:
            schema["input"] = []
            for field in self.inputs.values():
                schema["input"].append(field.to_schema())

        if self.outputs:
            schema["output"] = []
            for field in self.outputs.values():
                schema["output"].append(field.to_schema())

        schema["run"] = {
            "language": "python",
            "program": self.source or "",
        }

        return schema

    def __repr__(self):
        """Return string representation."""
        return "<ProcessDescriptor metadata={} inputs={} outputs={}>".format(
            repr(self.metadata),
            repr(dict(self.inputs)),
            repr(dict(self.outputs)),
        )



"""Resolwe custom serializer fields."""
from collections import OrderedDict

from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import smart_text

from rest_framework import exceptions, relations

from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.permissions.utils import get_full_perm


class DictRelatedField(relations.RelatedField):
    """
    Field representing the target of the relationship by using dict.

    A read-write field that represents the target of the relationship by
    a dictionary, where objects are uniquely defined by ``id`` or
    ``slug`` keys in the dictionary. If ``id`` is provided the uniquness
    of object is arbitrary. If ``id`` is not in the dict, object is
    determined by slug. Since multiple objects can have same slug, we
    filter objects by slug, by permissions and return object with
    highest version.
    """

    default_error_messages = {
        "slug_not_allowed": ("Use id (instead of slug) for update requests."),
        "null": ("At least one of id, slug must be present in field {name}."),
        "does_not_exist": (
            "Invalid {model_name} value: {value} - object does not exist."
        ),
        "permission_denied": (
            "You do not have {permission} permission for {model_name}: {value}."
        ),
    }

    def __init__(self, serializer, write_permission="view", **kwargs):
        """Initialize attributes."""
        self.serializer = serializer
        self.write_permission = write_permission
        super().__init__(**kwargs)

    @property
    def model_name(self):
        """Get name of queryset model."""
        return self.get_queryset().model._meta.model_name

    def to_internal_value(self, data):
        """Convert to internal value."""
        # Allow None values only in case field is not required.
        if "id" in data and data["id"] is None and not self.required:
            return
        elif data.get("id", None) is not None:
            kwargs = {"id": data["id"]}
        elif data.get("slug", None) is not None:
            if self.root.instance:
                # ``self.root.instance != None`` means that an instance is
                # already present, so this is not "create" request.
                self.fail("slug_not_allowed")
            kwargs = {"slug": data["slug"]}
        else:
            self.fail("null", name=self.field_name)

        user = getattr(self.context.get("request"), "user")
        queryset = self.get_queryset()
        permission = get_full_perm(self.write_permission, queryset.model)
        try:
            return get_objects_for_user(
                user, permission, queryset.filter(**kwargs)
            ).latest("version")
        except ObjectDoesNotExist:
            # Differentiate between "user has no permission" and "object does not exist"
            view_permission = get_full_perm("view", queryset.model)
            if permission != view_permission:
                try:
                    get_objects_for_user(
                        user, view_permission, queryset.filter(**kwargs)
                    ).latest("version")
                    raise exceptions.PermissionDenied(
                        "You do not have {} permission for {}: {}.".format(
                            self.write_permission, self.model_name, data
                        )
                    )
                except ObjectDoesNotExist:
                    pass

            self.fail(
                "does_not_exist", value=smart_text(data), model_name=self.model_name
            )

    def to_representation(self, obj):
        """Convert to representation."""
        serializer = self.serializer(obj, required=self.required, context=self.context)

        # Manually bind this serializer to field.parent as field projection
        # relies on parent/child relations between serializers.
        serializer.bind(self.field_name, self.parent)

        return serializer.data

    def get_choices(self, cutoff=None):
        """
        Get choices for a field.

        The caller of this method is BrowsableAPIRenderer (indirectly).

        Method is direct copy of ``RelatedField.get_choices``, with one
        modified line (noted below). Line is modified so ``OrderedDict``
        receives hashable keys. The impementation in parent class
        returns result of ``self.to_representaton``, which is not
        hashable in case of this field (``self.to_representaton``
        returns a ``ReturnDict`` object).
        """
        queryset = self.get_queryset()
        if queryset is None:
            # Ensure that field.choices returns something sensible
            # even when accessed with a read-only field.
            return {}

        if cutoff is not None:
            queryset = queryset[:cutoff]

        return OrderedDict(
            [
                (
                    # This line below is modifed.
                    item.pk,
                    self.display_value(item),
                )
                for item in queryset
            ]
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-04 16:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0010_add_secret"),
    ]

    operations = [
        migrations.AlterField(
            model_name="datadependency",
            name="parent",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="children_dependency",
                to="flow.Data",
            ),
        ),
    ]
